import pytest
import os
import glob as norm_glob
import XRootD.client.glob_funcs as glob
from pathlib2 import Path


@pytest.fixture
def tmptree(tmpdir):
    subdir1 = tmpdir / "subdir1"
    subdir1.mkdir()
    subdir2 = tmpdir / "subdir2"
    subdir2.mkdir()
    for i in range(3):
        dummy = subdir1 / ("a_file_%d.txt" % i)
        dummy.write_text(u"This is file %d\n" % i, encoding="utf-8")
    return tmpdir


def test_glob_local(tmptree):
    normal_glob_result = norm_glob.glob(str(tmptree / "not-there"))
    assert glob.glob(str(tmptree / "not-there")) == normal_glob_result
    assert len(glob.glob(str(tmptree / "not-there"))) == 0
    assert len(glob.glob(str(tmptree / "not-there*"))) == 0
    assert len(glob.glob(str(tmptree / "sub*"))) == 2
    assert len(glob.glob(str(tmptree / "subdir1" / "*txt"))) == 3
    assert len(glob.glob(str(tmptree / "subdir*" / "*txt"))) == 3

    with pytest.raises(RuntimeError) as excinfo:
        glob.glob(str(tmptree / "not-there"), raise_error=True)
    assert "[ERROR]" in str(excinfo.value)
    assert str(tmptree) in str(excinfo.value)


def test_glob_remote(tmptree):
    assert len(glob.glob("root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoad/")) == 0
    assert len(glob.glob("root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoa*")) == 1
    assert len(glob.glob("root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/*")) > 0
    assert len(glob.glob("root://eospublic.cern.ch//eos/root-*/cms_opendata_2012_nanoaod/*")) > 0

    with pytest.raises(RuntimeError) as excinfo:
        glob.glob("root://eospublic.cern.ch//eos/root-NOTREAL/cms_opendata_2012_nanoaod/*", raise_error=True)
    assert "[ERROR]" in str(excinfo.value)



#-------------------------------------------------------------------------------
# Copyright (c) 2012-2014 by European Organization for Nuclear Research (CERN)
# Author: Michal Simon <michal.simon@cern.ch>
#-------------------------------------------------------------------------------
# This file is part of the XRootD software suite.
#
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#------------------------------------------------------------------------------

from pyxrootd import client

__version__ = client.XrdVersion_cpp()


#-------------------------------------------------------------------------------
# Copyright (c) 2012-2014 by European Organization for Nuclear Research (CERN)
# Author: Michal Simon <michal.simon@cern.ch>
#-------------------------------------------------------------------------------
# This file is part of the XRootD software suite.
#
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#------------------------------------------------------------------------------

from pyxrootd import client


def EnvPutString( key, value ):
     """Sets the given key in the xrootd client environment to 
        the given value. Returns false if there is already a 
        shell-imported setting for this key, true otherwise
     """
     return client.EnvPutString_cpp( key, value )
 
def EnvGetString( key ):
     """Gets the given key from the xrootd client environment. 
        If key does not exist in the environment returns None.
     """
     return client.EnvGetString_cpp( key )
 
def EnvPutInt( key, value ):
     """Sets the given key in the xrootd client environment to 
        the given value. Returns false if there is already a 
        shell-imported setting for this key, true otherwise
     """
     return client.EnvPutInt_cpp( key, value )
 
def EnvGetInt( key ):
     """Gets the given key from the xrootd client environment. 
        If key does not exist in the environment returns None.
     """
     return client.EnvGetInt_cpp( key )


from XRootD import client
from XRootD.client.utils import AsyncResponseHandler
from XRootD.client.flags import OpenFlags, AccessMode
from env import *

import pytest
import sys
import os

# Global open mode
open_mode = (AccessMode.UR | AccessMode.UW |
             AccessMode.GR | AccessMode.GW |
             AccessMode.OR | AccessMode.OW)

def test_write_sync():
  f = client.File()
  pytest.raises(ValueError, "f.write(smallbuffer)")
  status, __ = f.open(smallfile, OpenFlags.DELETE, open_mode )
  assert status.ok

  # Write
  status, __ = f.write(smallbuffer)
  assert status.ok

  # Read
  status, response = f.read()
  assert status.ok
  assert len(response) == len(smallbuffer)

  buffer = 'eggs and ham\n'
  status, __ = f.write(buffer, offset=13, size=len(buffer) - 2)
  assert status.ok
  status, response = f.read()
  assert status.ok
  assert len(response) == len(buffer * 2) - 2
  f.close()

def test_write_async():
  f = client.File()
  status, __ = f.open(smallfile, OpenFlags.DELETE, open_mode)
  assert status.ok

  # Write async
  handler = AsyncResponseHandler()
  status = f.write(smallbuffer, callback=handler)
  status, __, __ = handler.wait()
  assert status.ok

  # Read sync
  status, response = f.read()
  assert status.ok
  assert len(response) == len(smallbuffer)
  f.close()

def test_open_close_sync():
  f = client.File()
  pytest.raises(ValueError, "f.stat()")
  status, __ = f.open(smallfile, OpenFlags.READ)
  assert status.ok
  assert f.is_open()

  # Close
  status, __ = f.close()
  assert status.ok
  assert f.is_open() == False
  pytest.raises(ValueError, "f.stat()")
  f.close()
  f.close()

def test_open_close_async():
  f = client.File()
  handler = AsyncResponseHandler()
  status = f.open(smallfile, OpenFlags.READ, callback=handler)
  assert status.ok
  status, __, __ = handler.wait()
  assert status.ok
  assert f.is_open()

  # Close async
  handler = AsyncResponseHandler()
  status = f.close(callback=handler)
  assert status.ok
  status, __, __ = handler.wait()
  assert status.ok
  assert f.is_open() == False

def test_io_limits():
  f = client.File()
  pytest.raises(ValueError, 'f.read()')
  status, __ = f.open(smallfile, OpenFlags.UPDATE)
  assert status.ok
  status, __ = f.stat()
  assert status.ok

  # Test read limits
  pytest.raises(TypeError, 'f.read(0, [1, 2])')
  pytest.raises(TypeError, 'f.read([1, 2], 0)')
  pytest.raises(TypeError, 'f.read(0, 10, [0, 1, 2])')
  pytest.raises(OverflowError, 'f.read(0, -10)')
  pytest.raises(OverflowError, 'f.read(-1, 1)')
  pytest.raises(OverflowError, 'f.read(0, 1, -1)')
  pytest.raises(OverflowError, 'f.read(0, 10**11)')
  pytest.raises(OverflowError, 'f.read(0, 10, 10**6)')

  # Test readline limits
  pytest.raises(TypeError, 'f.readline([0, 1], 1)')
  pytest.raises(TypeError, 'f.readline(0, [0, 1])')
  pytest.raises(TypeError, 'f.readline(0, 10, [0, 1])')
  pytest.raises(OverflowError, 'f.readline(-1, 1)')
  pytest.raises(OverflowError, 'f.readline(0, -1)')
  pytest.raises(OverflowError, 'f.readline(0, 1, -1)')
  pytest.raises(OverflowError, 'f.readline(0, 10**11)')
  pytest.raises(OverflowError, 'f.readline(0, 10, 10**11)')

  # Test write limits
  data = "data that will never get written"
  pytest.raises(TypeError, 'f.write(data, 0, [1, 2])')
  pytest.raises(TypeError, 'f.write(data, [1, 2], 0)')
  pytest.raises(TypeError, 'f.write(data, 0, 10, [0, 1, 2])')
  pytest.raises(OverflowError, 'f.write(data, 0, -10)')
  pytest.raises(OverflowError, 'f.write(data, -1, 1)')
  pytest.raises(OverflowError, 'f.write(data, 0, 1, -1)')
  pytest.raises(OverflowError, 'f.write(data, 0, 10**11)')
  pytest.raises(OverflowError, 'f.write(data, 0, 10, 10**6)')

  # Test vector_read limits
  pytest.raises(TypeError, 'f.vector_read(chunks=100)')
  pytest.raises(TypeError, 'f.vector_read(chunks=[1,2,3])')
  pytest.raises(TypeError, 'f.vector_read(chunks=[("lol", "cakes")])')
  pytest.raises(TypeError, 'f.vector_read(chunks=[(1), (2)])')
  pytest.raises(TypeError, 'f.vector_read(chunks=[(1, 2), (3)])')
  pytest.raises(OverflowError, 'f.vector_read(chunks=[(-1, -100), (-100, -100)])')
  pytest.raises(OverflowError, 'f.vector_read(chunks=[(0, 10**10*10)])')

  # Test truncate limits
  pytest.raises(TypeError, 'f.truncate(0, [1, 2])')
  pytest.raises(TypeError, 'f.truncate([1, 2], 0)')
  pytest.raises(OverflowError, 'f.truncate(-1)')
  pytest.raises(OverflowError, 'f.truncate(100, -10)')
  pytest.raises(OverflowError, 'f.truncate(0, 10**6)')
  status, __ = f.close()
  assert status.ok

def test_write_big_async():
  f = client.File()
  pytest.raises(ValueError, 'f.read()')
  status, __ = f.open(bigfile, OpenFlags.DELETE, open_mode)
  assert status.ok

  rand_data = os.urandom(64 * 1024)
  max_size = 512 * 1024  # 512 K
  offset = 0
  lst_handlers = []

  while offset <= max_size:
    status, __ = f.write(smallbuffer)
    assert status.ok
    handler = AsyncResponseHandler()
    lst_handlers.append(handler)
    status = f.write(rand_data, offset, callback=handler)
    assert status.ok
    offset = offset + len(smallbuffer) + len(rand_data)

  # Wait for async write responses
  for handler in lst_handlers:
    status, __, __ = handler.wait()
    assert status.ok

  f.close()

def test_read_sync():
  f = client.File()
  pytest.raises(ValueError, 'f.read()')
  status, response = f.open(bigfile, OpenFlags.READ)
  assert status.ok
  status, response = f.stat()
  size = response.size

  status, response = f.read()
  assert status.ok
  assert len(response) == size
  f.close()

def test_read_async():
  f = client.File()
  status, response = f.open(bigfile, OpenFlags.READ)
  assert status.ok
  status, response = f.stat()
  size = response.size

  handler = AsyncResponseHandler()
  status = f.read(callback=handler)
  assert status.ok
  status, response, hostlist = handler.wait()
  assert status.ok
  assert len(response) == size
  f.close()

def test_iter_small():
  f = client.File()
  status, __ = f.open(smallfile, OpenFlags.DELETE)
  assert status.ok
  status, __ = f.write(smallbuffer)
  assert status.ok

  size = f.stat(force=True)[1].size
  pylines = open('/tmp/spam').readlines()
  total = 0

  for i, line in enumerate(f):
    total += len(line)
    if pylines[i].endswith('\n'):
      assert line.endswith('\n')

  assert total == size
  f.close()

def test_iter_big():
  f = client.File()
  status, __ = f.open(bigfile, OpenFlags.READ)
  assert status.ok

  size = f.stat()[1].size
  pylines = open('/tmp/bigfile').readlines()
  total = 0

  for i, line in enumerate(f):
    total += len(line)
    if pylines[i].endswith('\n'):
      assert line.endswith('\n')

  assert total == size
  f.close()

def test_readline():
  f = client.File()
  f.open(smallfile, OpenFlags.DELETE, open_mode)
  f.write(smallbuffer)

  response = f.readline(offset=0, size=100)
  assert response == 'gre\0en\n'
  response = f.readline()
  assert response == 'eggs\n'
  response = f.readline()
  assert response == 'and\n'
  response = f.readline()
  assert response == 'ham\n'
  f.close()

  f = client.File()
  f.open(smallfile, OpenFlags.DELETE, open_mode)
  f.write(smallbuffer[:-1])
  f.readline()
  f.readline()
  f.readline()
  response = f.readline()
  assert response == 'ham'
  f.close()

def test_readlines_small():
  f = client.File()
  f.open(smallfile, OpenFlags.DELETE, open_mode)
  f.write(smallbuffer)
  f.close()
  pylines = open('/tmp/spam').readlines()

  for i in range(1, 100):
    f = client.File()
    f.open(smallfile)
    response = f.readlines(offset=0, chunksize=i)
    assert len(response) == 4
    for j, line in enumerate(response):
      if pylines[j].endswith('\n'):
        assert line.endswith('\n')
    f.close()

def test_readlines_big():
  f = client.File()
  f.open(bigfile, OpenFlags.READ)
  size = f.stat()[1].size

  lines = f.readlines()
  pylines = open('/tmp/bigfile').readlines()
  assert len(lines) == len(pylines)

  nlines = len(pylines)

  total = 0
  for i, l in enumerate(lines):
    total += len(l)
    if l != pylines[i]:
      print '!!!!!', total, i
      print '+++++ py: %r' % pylines[i]
      print '+++++ me: %r' % l
      break
    if pylines[i].endswith('\n'):
      assert l.endswith('\n')

  assert total == size
  f.close()

def test_readchunks_small():
  f = client.File()
  f.open(smallfile, OpenFlags.READ)
  size = f.stat()[1].size

  total = 0
  chunks = ['gre', '\0en', '\neg', 'gs\n', 'and', '\nha', 'm\n']
  for i, chunk in enumerate(f.readchunks(chunksize=3)):
    assert chunk == chunks[i]
    total += len(chunk)

  assert total == size
  f.close()

def test_readchunks_big():
  f = client.File()
  f.open(bigfile, OpenFlags.READ)
  size = f.stat()[1].size

  total = 0
  for chunk in f.readchunks(chunksize=1024 * 1024 * 2):
    total += len(chunk)

  assert total == size
  f.close()

def test_vector_read_sync():
  v = [(0, 100), (101, 200), (201, 200)]
  vlen = sum([vec[1] for vec in v])

  f = client.File()
  status, __ = f.open(bigfile, OpenFlags.READ)
  assert status.ok
  status, stat_info = f.stat()
  assert status.ok
  status, response = f.vector_read(chunks=v)

  # If big enough file everything shoud be ok
  if (stat_info.size > max([off + sz for (off, sz) in v])):
    assert status.ok
    assert response.size == vlen
  else:
    # If file not big enough this should fail
    status, __ = f.vector_read(chunks=v)
    assert not status.ok

  f.close()

def test_vector_read_async():
  v = [(0, 100), (101, 200), (201, 200)]
  vlen = sum([vec[1] for vec in v])
  f = client.File()
  status, __ = f.open(bigfile, OpenFlags.READ)
  assert status.ok
  status, stat_info = f.stat()
  assert status.ok
  handler = AsyncResponseHandler()
  status = f.vector_read(chunks=v, callback=handler)
  assert status.ok

  # If big enough file everything shoud be ok
  if (stat_info.size > max([off + sz for (off, sz) in v])):
    status, response, hostlist = handler.wait()
    assert status.ok
    assert response.size == vlen
  else:
    status, __, __ = handler.wait()
    assert not status.ok

  f.close()

def test_stat_sync():
  f = client.File()
  pytest.raises(ValueError, 'f.stat()')
  status, __ = f.open(bigfile)
  assert status.ok
  status, __ = f.stat()
  assert status.ok
  f.close()

def test_stat_async():
  f = client.File()
  status, response = f.open(bigfile)
  assert status.ok

  handler = AsyncResponseHandler()
  status = f.stat(callback=handler)
  assert status.ok
  status, __, __ = handler.wait()
  assert status.ok
  f.close()

def test_sync_sync():
  f = client.File()
  pytest.raises(ValueError, 'f.sync()')
  status, __ = f.open(bigfile)
  assert status.ok
  status, __ = f.sync()
  assert status.ok
  f.close()

def test_sync_async():
  f = client.File()
  status, response = f.open(bigfile)
  assert status.ok

  handler = AsyncResponseHandler()
  status = f.sync(callback=handler)
  status, __, __ = handler.wait()
  assert status.ok
  f.close()

def test_truncate_sync():
  f = client.File()
  pytest.raises(ValueError, 'f.truncate(10000)')
  status, __ = f.open(smallfile, OpenFlags.DELETE)
  assert status.ok

  status, __ = f.truncate(size=10000)
  assert status.ok
  f.close()

def test_truncate_async():
  f = client.File()
  status, __ = f.open(smallfile, OpenFlags.DELETE)
  assert status.ok

  handler = AsyncResponseHandler()
  status = f.truncate(size=10000, callback=handler)
  assert status.ok
  status, __, __ = handler.wait()
  assert status.ok
  f.close()

def test_misc():
  f = client.File()
  assert not f.is_open()

  # Open
  status, response = f.open(smallfile, OpenFlags.READ)
  assert status.ok
  assert f.is_open()

  # Set/get file properties
  f.set_property("ReadRecovery", "true")
  f.set_property("WriteRecovery", "true")
  assert f.get_property("DataServer")

  # Testing context manager
  f.close()
  assert not f.is_open()




from XRootD import client
from XRootD.client.flags import OpenFlags
from time import sleep

def callback( status, response, hostlist ):
  print "Called:", status, response, hostlist

with client.File() as f:
  status, response = f.open('root://localhost//tmp/eggs', OpenFlags.DELETE)
  status = f.fcntl( 'asdf', callback = callback )
  sleep(20)



from XRootD import client
from XRootD.client.flags import OpenFlags
from threading import Thread
from env import *

class TestThread(Thread):
  def __init__(self, file, id):
    Thread.__init__(self)
    self.file = file
    self.id = id
  def run(self):
    self.file.open(smallfile, OpenFlags.DELETE)
    assert self.file.is_open()

    s, _ = self.file.write(smallbuffer)
    assert s.ok

    print '+++ thread %d says: %s' % (self.id, self.file.read())

    for line in self.file:
      print '+++ thread %d says: %s' % (self.id, line)

    self.file.close()

def test_threads():
  f = client.File()
#   f.open(smallfile, OpenFlags.DELETE)
#   assert f.is_open()
#   f.write(smallbuffer)

  for i in xrange(3):
    tt = TestThread(f, i)
    tt.start()
    tt.join()

#   f.close()



"""
Copy a file
-----------

See :mod:`XRootD.client.CopyProcess` if you need multiple/more configurable
copy jobs.
"""
from XRootD import client

myclient = client.FileSystem('root://localhost')
status = myclient.copy('/tmp/spam', '/tmp/eggs', force=True)
assert status[0].ok



#-------------------------------------------------------------------------------
# Copyright (c) 2012-2014 by European Organization for Nuclear Research (CERN)
# Author: Justin Salmon <jsalmon@cern.ch>
#-------------------------------------------------------------------------------
# This file is part of the XRootD software suite.
#
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#-------------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function

from pyxrootd import client
from XRootD.client.responses import XRootDStatus, StatInfo, VectorReadInfo
from XRootD.client.utils import CallbackWrapper

class File(object):
  """Interact with an ``xrootd`` server to perform file-based operations such
  as reading, writing, vector reading, etc."""

  def __init__(self):
    self.__file = client.File()

  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    self.__file.__exit__()

  def __iter__(self):
    return self

  def __next__(self):
    return self.__file.next()

  # Python 2 compatibility
  next = __next__

  def open(self, url, flags=0, mode=0, timeout=0, callback=None):
    """Open the file pointed to by the given URL.

    :param   url: url of the file to be opened
    :type    url: string
    :param flags: An `ORed` combination of :mod:`XRootD.client.flags.OpenFlags`
                  where the default is `OpenFlags.NONE`
    :param  mode: access mode for new files, an `ORed` combination of
                 :mod:`XRootD.client.flags.AccessMode` where the default is
                 `AccessMode.NONE`
    :returns:    tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                 object and None
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.open(url, flags, mode, timeout, callback))

    status, response = self.__file.open(url, flags, mode, timeout)
    return XRootDStatus(status), None

  def close(self, timeout=0, callback=None):
    """Close the file.

    :returns: tuple containing :mod:`XRootD.client.responses.XRootDStatus`
              object and None

    As of Python 2.5, you can avoid having to call this method explicitly if you
    use the :keyword:`with` statement.  For example, the following code will
    automatically close *f* when the :keyword:`with` block is exited::

      from __future__ import with_statement # This isn't required in Python 2.6

      with client.File() as f:
        f.open("root://someserver//somefile")
        for line in f:
          print line,
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.close(timeout, callback))

    status, response = self.__file.close(timeout)
    return XRootDStatus(status), None

  def stat(self, force=False, timeout=0, callback=None):
    """Obtain status information for this file.

    :param force: do not use the cached information, force re-stating
    :type  force: boolean
    :returns:     tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                  object and :mod:`XRootD.client.responses.StatInfo` object
    """
    if callback:
      callback = CallbackWrapper(callback, StatInfo)
      return XRootDStatus(self.__file.stat(force, timeout, callback))

    status, response = self.__file.stat(force, timeout)
    if response: response = StatInfo(response)
    return XRootDStatus(status), response

  def read(self, offset=0, size=0, timeout=0, callback=None):
    """Read a data chunk from a given offset.

    :param offset: offset from the beginning of the file
    :type  offset: integer
    :param   size: number of bytes to be read
    :type    size: integer
    :returns:      tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                   object and the data that was read
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.read(offset, size, timeout, callback))

    status, response = self.__file.read(offset, size, timeout)
    return XRootDStatus(status), response

  def readline(self, offset=0, size=0, chunksize=0):
    """Read a data chunk from a given offset, until the first newline or EOF
    encountered.

    :param    offset: offset from the beginning of the file
    :type     offset: integer
    :param      size: maximum number of bytes to be read
    :type       size: integer
    :param chunksize: size of chunk used for reading, in bytes
    :type  chunksize: integer
    :returns:         data that was read, including the trailing newline
    :rtype:           string
    """
    return self.__file.readline(offset, size, chunksize)

  def readlines(self, offset=0, size=0, chunksize=0):
    """Read lines from a given offset until EOF encountered. Return list of
    lines read.

    :param    offset: offset from the beginning of the file
    :type     offset: integer
    :param      size: maximum number of bytes to be read
    :type       size: integer
    :param chunksize: size of chunk used for reading, in bytes
    :type  chunksize: integer
    :returns:         data that was read, including trailing newlines
    :rtype:           list of strings

    .. warning:: This method will read the whole file into memory if you don't
                 specify an offset. Think twice about using it if your files
                 are big.
    """
    return self.__file.readlines(offset, size, chunksize)

  def readchunks(self, offset=0, chunksize=1024 * 1024 * 2):
    """Return an iterator object which will read data chunks from a given
    offset of the given chunksize until EOF.

    :param    offset: offset from the beginning of the file
    :type     offset: integer
    :param chunksize: size of chunk to read, in bytes
    :type  chunksize: integer
    :returns:         iterator object
    """
    return self.__file.readchunks(offset, chunksize)

  def write(self, buffer, offset=0, size=0, timeout=0, callback=None):
    """Write a data chunk at a given offset.

    :param buffer: data to be written
    :param offset: offset from the beginning of the file
    :type  offset: integer
    :param   size: number of bytes to be written
    :type    size: integer
    :returns:      tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                   object and None
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.write(buffer, offset, size, timeout, callback))

    status, response = self.__file.write(buffer, offset, size, timeout)
    return XRootDStatus(status), None

  def sync(self, timeout=0, callback=None):
    """Commit all pending disk writes.

    :returns: tuple containing :mod:`XRootD.client.responses.XRootDStatus`
              object and None
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.sync(timeout, callback))

    status, response = self.__file.sync(timeout)
    return XRootDStatus(status), None

  def truncate(self, size, timeout=0, callback=None):
    """Truncate the file to a particular size.

    :param size: desired size of the file
    :type  size: integer
    :returns:    tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                 object and None
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.truncate(size, timeout, callback))

    status, response = self.__file.truncate(size, timeout)
    return XRootDStatus(status), None

  def vector_read(self, chunks, timeout=0, callback=None):
    """Read scattered data chunks in one operation.

    :param chunks: list of the chunks to be read. The default maximum
                   chunk size is 2097136 bytes and the default maximum
                   number of chunks per request is 1024. The server may
                   be queried using :func:`XRootD.client.FileSystem.query`
                   for the actual settings.
    :type  chunks: list of 2-tuples of the form (offset, size)
    :returns:      tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                   object and :mod:`XRootD.client.responses.VectorReadInfo`
                   object
    """
    if callback:
      callback = CallbackWrapper(callback, VectorReadInfo)
      return XRootDStatus(self.__file.vector_read(chunks, timeout, callback))

    status, response = self.__file.vector_read(chunks, timeout)
    if response: response = VectorReadInfo(response)
    return XRootDStatus(status), response

  def fcntl(self, arg, timeout=0, callback=None):
    """Perform a custom operation on an open file.

    :param    arg: argument
    :type     arg: string
    :returns:      tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                   object and a string
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.fcntl( arg, timeout, callback))

    status, response = self.__file.fcntl( arg, timeout )
    return XRootDStatus(status), response

  def visa(self, timeout=0, callback=None):
    """Get access token to a file.

    :returns:      tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                   object and a string
    """
    if callback:
      callback = CallbackWrapper(callback, None)
      return XRootDStatus(self.__file.visa(timeout, callback))

    status, response = self.__file.visa(timeout)
    return XRootDStatus(status), response

  def is_open(self):
    """Check if the file is open.

    :rtype: boolean
    """
    return self.__file.is_open()

  def set_property(self, name, value):
    """Set file property.

    :param name: name of the property to set
    :type  name: string
    :returns:    boolean denoting if property setting was successful
    :rtype:      boolean
    """
    return self.__file.set_property(name, value)

  def get_property(self, name):
    """Get file property.

    :param name: name of the property
    :type  name: string
    """
    return self.__file.get_property(name)

  def set_xattr(self, attrs, timeout=0, callback=None):
    """Set extended file attributes.

    :param attrs: extended attributes to be set on the file
    :type  attrs: list of tuples of name/value pairs
    :returns:     tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                  object and :mod:`list of touples (string, XRootD.client.responses.XRootDStatus)` object
    """
    if callback:
      callback = CallbackWrapper(callback, list)
      return XRootDStatus(self.__file.set_xattr(attrs, timeout, callback))

    status, response = self.__file.set_xattr(attrs, timeout)
    return XRootDStatus(status), response

  def get_xattr(self, attrs, timeout=0, callback=None):
    """Get extended file attributes.

    :param attrs: list of extended attribute names to be retrived
    :type  attrs: list of strings
    :returns:     tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                  object and :mod:`list of touples (string, string, XRootD.client.responses.XRootDStatus)` object
    """
    if callback:
      callback = CallbackWrapper(callback, list)
      return XRootDStatus(self.__file.get_xattr(attrs, timeout, callback))

    status, response = self.__file.get_xattr(attrs, timeout)
    return XRootDStatus(status), response

  def del_xattr(self, attrs, timeout=0, callback=None):
    """Delete extended file attributes.

    :param attrs: list of extended attribute names to be deleted
    :type  attrs: list of strings
    :returns:     tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                  object and :mod:`list of touples (string, XRootD.client.responses.XRootDStatus)` object
    """
    if callback:
      callback = CallbackWrapper(callback, list)
      return XRootDStatus(self.__file.del_xattr(attrs, timeout, callback))

    status, response = self.__file.del_xattr(attrs, timeout)
    return XRootDStatus(status), response

  def list_xattr(self, timeout=0, callback=None):
    """List all extended file attributes.

    :returns:     tuple containing :mod:`XRootD.client.responses.XRootDStatus`
                  object and :mod:`list of touples (string, string, XRootD.client.responses.XRootDStatus)` object
    """
    if callback:
      callback = CallbackWrapper(callback, list)
      return XRootDStatus(self.__file.list_xattr(timeout, callback))

    status, response = self.__file.list_xattr(timeout)
    return XRootDStatus(status), response




#-------------------------------------------------------------------------------
# Copyright (c) 2012-2013 by European Organization for Nuclear Research (CERN)
# Author: Justin Salmon <jsalmon@cern.ch>
#-------------------------------------------------------------------------------
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http:#www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function

from XRootD.client.url import URL

class Struct(object):
  """Convert a dict into an object by adding each dict entry to __dict__"""
  def __init__(self, entries):
      self.__dict__.update(**entries)
  def __repr__(self):
    return '<%s>' % str(', '.join('%s: %s' % (k, repr(v))
                                  for (k, v) in self.__dict__.items()))

class LocationInfo(Struct):
  """Path location information (a list of discovered file locations).

  :param locations: (List of :mod:`XRootD.client.responses.Location` objects)
                    List of discovered locations

  This object is iterable::

    >>> status, locations = filesystem.locate('/tmp', OpenFlags.REFRESH)
    >>> print locations
    <XRootD.client.responses.LocationInfo object at 0x288b9f0>
    >>> for location in locations:
    ...   print location.address
    ...
    [::127.0.0.1]:1094

  """
  def __init__(self, locations):
    super(LocationInfo, self).__init__({'locations':
                                        [Location(l) for l in locations]})

  def __iter__(self):
    return iter(self.locations)

class Location(Struct):
  """Information about a single location.

  :var    address: The address of this location
  :var       type: The type of this location, one of
                   :mod:`XRootD.client.flags.LocationType`
  :var accesstype: The allowed access type of this location, one of
                   :mod:`XRootD.client.flags.AccessType`
  :var is_manager: Is the location a manager
  :var  is_server: Is the location a server
  """
  def __init__(self, location):
    super(Location, self).__init__(location)

class XRootDStatus(Struct):
  """Status of a request. Returned with all requests.

  :var   message: Message describing the status of this request
  :var        ok: The request was successful
  :var     error: Error making request
  :var     fatal: Fatal error making request
  :var    status: Status of the request
  :var      code: Error type, or additional hints on what to do
  :var shellcode: Status code that may be returned to the shell
  :var     errno: Errno, if any
  """

  #----------------------------------------------------------------------------
  # Additional info for the stOK status
  #----------------------------------------------------------------------------
  suDone            = 0
  suContinue        = 1
  suRetry           = 2
  suPartial         = 3
  suAlreadyDone     = 4

  #----------------------------------------------------------------------------
  # Generic errors
  #----------------------------------------------------------------------------
  errNone           = 0; # No error
  errRetry          = 1; # Try again for whatever reason
  errUnknown        = 2; # Unknown error
  errInvalidOp      = 3; # The operation cannot be performed in the
                         # given circumstances
  errFcntl          = 4; # failed manipulate file descriptor
  errPoll           = 5; # error while polling descriptors
  errConfig         = 6; # System misconfigured
  errInternal       = 7; # Internal error
  errUnknownCommand = 8;
  errInvalidArgs    = 9;
  errInProgress     = 10;
  errUninitialized  = 11;
  errOSError        = 12;
  errNotSupported   = 13;
  errDataError      = 14; # data is corrupted
  errNotImplemented = 15; # Operation is not implemented
  errNoMoreReplicas = 16; # No more replicas to try
  errPipelineFailed = 17; # Pipeline failed and operation couldn't be executed

  #----------------------------------------------------------------------------
  # Socket related errors
  #----------------------------------------------------------------------------
  errInvalidAddr        = 101;
  errSocketError        = 102;
  errSocketTimeout      = 103;
  errSocketDisconnected = 104;
  errPollerError        = 105;
  errSocketOptError     = 106;
  errStreamDisconnect   = 107;
  errConnectionError    = 108;
  errInvalidSession     = 109;

  #----------------------------------------------------------------------------
  # Post Master related errors
  #----------------------------------------------------------------------------
  errInvalidMessage       = 201;
  errHandShakeFailed      = 202;
  errLoginFailed          = 203;
  errAuthFailed           = 204;
  errQueryNotSupported    = 205;
  errOperationExpired     = 206;
  errOperationInterrupted = 207;

  #----------------------------------------------------------------------------
  # XRootD related errors
  #----------------------------------------------------------------------------
  errNoMoreFreeSIDs     = 301;
  errInvalidRedirectURL = 302;
  errInvalidResponse    = 303;
  errNotFound           = 304;
  errCheckSumError      = 305;
  errRedirectLimit      = 306;

  errErrorResponse      = 400;
  errRedirect           = 401;

  errResponseNegative   = 500; # Query response was negative

  def __init__(self, status):
    super(XRootDStatus, self).__init__(status)

  def __str__(self):
    return self.message

class ProtocolInfo(Struct):
  """Protocol information for a server.

  :var  version: The version of the protocol this server is speaking
  :var hostinfo: Informational flags for this host. An `ORed` combination of
                 :mod:`XRootD.client.flags.HostTypes`
  """
  def __init__(self, info):
    super(ProtocolInfo, self).__init__(info)

class StatInfo(Struct):
  """Status information for files and directories.

  :var         id: This file's unique identifier
  :var      flags: Informational flags. An `ORed` combination of
                   :mod:`XRootD.client.flags.StatInfoFlags`
  :var       size: The file size (in bytes)
  :var    modtime: Modification time (in seconds since epoch)
  :var modtimestr: Modification time (as readable string)
  """
  def __init__(self, info):
    super(StatInfo, self).__init__(info)

class StatInfoVFS(Struct):
  """Status information for Virtual File Systems.

  :var            nodes_rw: Number of nodes that can provide read/write space
  :var             free_rw: Size of the largest contiguous area of free r/w
                            space (in MB)
  :var      utilization_rw: Percentage of the partition utilization represented
                            by ``free_rw``
  :var       nodes_staging: Number of nodes that can provide staging space
  :var        free_staging: Size of the largest contiguous area of free staging
                            space (in MB)
  :var utilization_staging: Percentage of the partition utilization represented
                            by ``free_staging``
  """
  def __init__(self, info):
    super(StatInfoVFS, self).__init__(info)

class DirectoryList(Struct):
  """Directory listing.

  This object is iterable::

    >>> status, dirlist = filesystem.dirlist('/tmp', DirListFlags.STAT)
    >>> print dirlist
    <XRootD.client.responses.DirectoryList object at 0x288b9f0>
    >>> print 'Entries:', dirlist.size
    Entries: 2
    >>> for item in dirlist:
    ...   print item.name, item.statinfo.size
    ...
    spam 1024
    eggs 2048

  :var     size: The size of this listing (number of entries)
  :var   parent: The name of the parent directory of this directory
  :var  dirlist: (List of :mod:`XRootD.client.responses.ListEntry` objects) -
                 The list of directory entries
  """
  def __init__(self, dirlist):
    dirlist.update({'dirlist': [ListEntry(e) for e in dirlist['dirlist']]})
    super(DirectoryList, self).__init__(dirlist)

  def __iter__(self):
    return iter(self.dirlist)

class ListEntry(Struct):
  """An entry in a directory listing.

  :var      name: The name of the file/directory
  :var  hostaddr: The address of the host on which this file/directory lives
  :var  statinfo: (Instance of :mod:`XRootD.client.responses.StatInfo`) -
                  Status information about this file/directory. You must pass
                  `DirListFlags.STAT` with the call to
                  :mod:`XRootD.client.FileSystem.dirlist()` to retrieve status
                  information.
  """
  def __init__(self, entry):
    if entry['statinfo']: entry.update({'statinfo': StatInfo(entry['statinfo'])})
    super(ListEntry, self).__init__(entry)

class ChunkInfo(Struct):
  """Describes a data chunk for a vector read.

  :var offset: The offset in the file from which this chunk came
  :var length: The length of this chunk
  :var buffer: The actual chunk data
  """
  def __init__(self, info):
    super(ChunkInfo, self).__init__(info)

class VectorReadInfo(Struct):
  """Vector read response object.
  Returned by :mod:`XRootD.client.File.vector_read()`.

  This object is iterable::

    >>> f.open('root://localhost/tmp/spam')
    >>> status, chunks = file.vector_read([(0, 10), (10, 10)])
    >>> print chunks
    <XRootD.client.responses.VectorReadInfo object at 0x288b9f0>
    >>> print chunks.size
    20
    >>> for chunk in chunks:
    ...   print chunk.offset, chunk.length
    ...
    0 10
    10 10

  :var    size: Total size of all chunks
  :var  chunks: (List of :mod:`XRootD.client.responses.ChunkInfo` objects) -
                The list of chunks that were read
  """
  def __init__(self, info):
    info.update({'chunks': [ChunkInfo(c) for c in info['chunks']]})
    super(VectorReadInfo, self).__init__(info)

  def __iter__(self):
    return iter(self.chunks)

class HostList(Struct):
  """A list of hosts that were involved in the request.

  This object is iterable::

    >>> print hostlist
    <XRootD.client.responses.HostList object at 0x288b9f0>
    >>> for host in hostlist:
    ...   print host.url
    ...
    root://localhost

  :var  hosts: (List of :mod:`XRootD.client.responses.HostInfo` objects) -
               The list of hosts
  """
  def __init__(self, hostlist):
    super(HostList, self).__init__({'hosts': [HostInfo(h) for h in hostlist]})

  def __iter__(self):
    return iter(self.hosts)

class HostInfo(Struct):
  """Information about a single host.

  :var           url: URL of the host, instance of :mod:`XRootD.client.URL`
  :var      protocol: Version of the protocol the host is speaking
  :var         flags: Host type, an `ORed` combination of
                      :mod:`XRootD.client.flags.HostTypes`
  :var load_balancer: Was the host used as a load balancer
  """
  def __init__(self, info):
    super(HostInfo, self).__init__(info)



from setuptools import setup, Extension
from setuptools.command.install import install
from setuptools.command.sdist import sdist
from wheel.bdist_wheel import bdist_wheel

import subprocess
import sys
import getpass

def get_version():
    version = subprocess.check_output(['./genversion.sh', '--print-only'])
    version = version.decode()
    if version.startswith('v'):
        version = version[1:]
    version = version.split('-')[0]
    return version

def get_version_from_file():
    try:
        f = open('./bindings/python/VERSION')
        version = f.read().split('/n')[0]
        f.close()
        return version
    except:
        print('Failed to get version from file. Using unknown')
        return 'unknown'

def binary_exists(name):
    """Check whether `name` is on PATH."""
    from distutils.spawn import find_executable
    return find_executable(name) is not None

def check_cmake3(path):
    from distutils.spawn import find_executable
    args = (path, "--version")
    popen = subprocess.Popen(args, stdout=subprocess.PIPE)
    popen.wait()
    output = popen.stdout.read().decode("utf-8") 
    prefix_len = len( "cmake version " )
    version = output[prefix_len:].split( '.' )
    return int( version[0] ) >= 3

def cmake_exists():
    """Check whether CMAKE is on PATH."""
    from distutils.spawn import find_executable
    path = find_executable('cmake')
    if path is not None:
        if check_cmake3(path): return True, path
    path = find_executable('cmake3')
    return path is not None, path 



# def python_dependency_name( py_version_short, py_version_nodot ):
#     """ find the name of python dependency """
#     from distutils.spawn import find_executable
#     # this is the path to default python
#     path = find_executable( 'python' )
#     from os.path import realpath
#     # this is the real default python after resolving symlinks
#     real = realpath(path)
#     index = real.find( 'python' ) + len( 'python' )
#     # this is the version of default python
#     defaultpy = real[index:]
#     if defaultpy != py_version_short:
#       return 'python' + py_version_nodot
#     return 'python'

class CustomInstall(install):
    def run(self): 

        py_version_short = self.config_vars['py_version_short']
        py_version_nodot = self.config_vars['py_version_nodot']

        cmake_bin, cmake_path = cmake_exists()
        make_bin    = binary_exists( 'make' )
        comp_bin    = binary_exists( 'c++' ) or binary_exists( 'g++' ) or binary_exists( 'clang' )

        import pkgconfig
        zlib_dev    = pkgconfig.exists( 'zlib' )
        openssl_dev = pkgconfig.exists( 'openssl' )
        uuid_dev    = pkgconfig.exists( 'uuid' )
        
        pyname = None
        if py_version_nodot[0] == '3':
            python_dev = pkgconfig.exists( 'python3' ) or pkgconfig.exists( 'python' + py_version_nodot );
            pyname = 'python3'
        else:
            python_dev = pkgconfig.exists( 'python' );
            pyname = 'python'

        missing_dep = not ( cmake_bin and make_bin and comp_bin and zlib_dev and openssl_dev and python_dev and uuid_dev )

        if missing_dep:
          print( 'Some dependencies are missing:')
          if not cmake_bin:   print('\tcmake (version 3) is missing!')
          if not make_bin:    print('\tmake is missing!')
          if not comp_bin:    print('\tC++ compiler is missing (g++, c++, clang, etc.)!')
          if not zlib_dev:    print('\tzlib development package is missing!')
          if not openssl_dev: print('\topenssl development package is missing!')
          if not python_dev:  print('\t{} development package is missing!'.format(pyname) )
          if not uuid_dev:    print('\tuuid development package is missing')
          raise Exception( 'Dependencies missing!' )

        useropt = ''
        command = ['./install.sh']
        if self.user:
            prefix = self.install_usersite
            useropt = '--user'
        else:
            prefix = self.install_platlib
        command.append(prefix)
        command.append( py_version_short )
        command.append( useropt )
        command.append( cmake_path )
        rc = subprocess.call(command)
        if rc:
          raise Exception( 'Install step failed!' )


class CustomDist(sdist):
    def write_version_to_file(self):

        version = get_version()
        with open('bindings/python/VERSION', 'w') as vi:
            vi.write(version)
    
    def run(self):
        self.write_version_to_file()
        sdist.run(self)


class CustomWheelGen(bdist_wheel):
    # Do not generate wheel
    def run(self):
        pass


version = get_version()
if version.startswith('unknown'):
    version = get_version_from_file()

setup_requires=[ 'pkgconfig' ]

setup( 
    name             = 'xrootd',
    version          = version,
    author           = 'XRootD Developers',
    author_email     = 'xrootd-dev@slac.stanford.edu',
    url              = 'http://xrootd.org',
    license          = 'LGPLv3+',
    description      = "XRootD with Python bindings",
    long_description = "XRootD with Python bindings",
    setup_requires   = setup_requires,
    cmdclass         = {
        'install':     CustomInstall,
        'sdist':       CustomDist,
        'bdist_wheel': CustomWheelGen
    }
)



from XRootD import client
from XRootD.client.flags import OpenFlags
from env import *

def test_copy_smallfile():

  f = client.File()
  s, r = f.open(smallfile, OpenFlags.DELETE )
  assert s.ok
  f.write(smallbuffer)
  size1 = f.stat(force=True)[1].size
  f.close()

  c = client.CopyProcess()
  c.add_job( source=smallfile, target=smallcopy, force=True )
  s = c.prepare()
  assert s.ok
  s, __ = c.run()
  assert s.ok

  f = client.File()
  s, r = f.open(smallcopy, OpenFlags.READ)
  size2 = f.stat()[1].size

  assert size1 == size2
  f.close()

def test_copy_bigfile():
  f = client.File()
  s, r = f.open(bigfile)
  assert s.ok
  size1 = f.stat(force=True)[1].size
  f.close()

  c = client.CopyProcess()
  c.add_job( source=bigfile, target=bigcopy, force=True )
  s = c.prepare()
  assert s.ok
  s, __ = c.run()
  assert s.ok

  f = client.File()
  s, r = f.open(bigcopy, OpenFlags.READ)
  size2 = f.stat()[1].size

  assert size1 == size2
  f.close()

def test_copy_nojobs():
  c = client.CopyProcess()
  s = c.prepare()
  assert s.ok
  s, __ = c.run()
  assert s.ok

def test_copy_noprep():
  c = client.CopyProcess()
  c.add_job( source=bigfile, target=bigcopy, force=True )
  s, __ = c.run()
  assert s.ok

class TestProgressHandler(object):
  def begin(self, id, total, source, target):
    print '+++ begin(): %d, total: %d' % (id, total)
    print '+++ source: %s' % source
    print '+++ target: %s' % target

  def end(self, jobId, status):
    print '+++ end(): jobId: %s, status: %s'  % (jobId, status)

  def update(self, jobId, processed, total):
    print '+++ update(): jobid: %s, processed: %d, total: %d' % (jobId, processed, total)

def test_copy_progress_handler():
  c = client.CopyProcess()
  c.add_job( source=bigfile, target=bigcopy, force=True )
  c.prepare()

  h = TestProgressHandler()
  c.run(handler=h)



#-------------------------------------------------------------------------------
# Copyright (c) 2012-2014 by European Organization for Nuclear Research (CERN)
# Author: Michal Simon <michal.simon@cern.ch>
#-------------------------------------------------------------------------------
# This file is part of the XRootD software suite.
#
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#------------------------------------------------------------------------------
from __future__ import absolute_import

import gc
import atexit

from pyxrootd import client
from .file import File


@atexit.register
def finalize():
     """Python atexit handler, will stop all XRootD client threads
        (XrdCl JobManager, TaskManager and Poller) in order to ensure 
        no Python APIs are called after the Python Interpreter gets
        finalized.
     """
     # Ensure there are no files left open as calling their destructor will
     # cause "close" commands to be sent.
     # If this isn't done, there will be no running threads to process requests
     # after this function returns so the interpreter will deadlock.
     for obj in gc.get_objects():
         if isinstance(obj, File) and obj.is_open():
             obj.close()

     client.__XrdCl_Stop_Threads()



"""
Read a certain amount of data from a certain offset in a file
-------------------------------------------------------------

Produces the following output::

  'green\neggs\nand\nham\n'
  'eggs'

"""
from XRootD import client
from XRootD.client.flags import OpenFlags

with client.File() as f:
  status, response = f.open('root://localhost//tmp/eggs', OpenFlags.DELETE)
  f.write('green\neggs\nand\nham\n')

  status, data = f.read() # Reads the whole file
  print '%r' % data
  print f.get_property('DataServer')
  print f.get_property('LastURL')
  print f.get_property('ReadRecovery')
  f.set_property('ReadRecovery', 'false')
  print f.get_property('ReadRecovery')

  status, data = f.read(offset=6, size=4) # Reads "eggs"
  print '%r' % data


"""
Locate a file
-------------

Produces output similar to the following::

  <locations: [<type: 2, address: '[::127.0.0.1]:1094', accesstype: 1, is_manager: False, is_server: True>]>
"""
from XRootD import client
from XRootD.client.flags import OpenFlags

myclient = client.FileSystem("root://localhost")
status, locations = myclient.locate("/tmp", OpenFlags.REFRESH)

print locations



"""
Make a directory, remove it and test the filesystem properties
--------------------------------------------------------------
"""
from XRootD import client
from XRootD.client.flags import MkDirFlags

myclient = client.FileSystem("root://localhost")
myclient.mkdir("/tmp/some/dir", MkDirFlags.MAKEPATH)
myclient.rmdir("/tmp/some/dir")
myclient.rmdir("/tmp/some")

print myclient.get_property( "FollowRedirects" )
print myclient.set_property( "FollowRedirects", "false" )
print myclient.get_property( "FollowRedirects" )




from XRootD import client
from XRootD.client.utils import AsyncResponseHandler
from XRootD.client.flags import OpenFlags, QueryCode, MkDirFlags, AccessMode, \
                                 DirListFlags, PrepareFlags
from env import *
import pytest
import sys
import os
import inspect

def test_filesystem():
  c = client.FileSystem(SERVER_URL)

  funcspecs = [(c.locate,     ('/tmp', OpenFlags.REFRESH), True),
               (c.deeplocate, ('/tmp', OpenFlags.REFRESH), True),
               (c.query,      (QueryCode.SPACE, '/tmp'), True),
               (c.truncate,   ('/tmp/spam', 1000), False),
               (c.mv,         ('/tmp/spam', '/tmp/ham'), False),
               (c.chmod,      ('/tmp/ham', AccessMode.UR | AccessMode.UW), False),
               (c.rm,         ('/tmp/ham',), False),
               (c.mkdir,      ('/tmp/somedir', MkDirFlags.MAKEPATH), False),
               (c.rmdir,      ('/tmp/somedir',), False),
               (c.ping,       (), False),
               (c.stat,       ('/tmp',), True),
               (c.statvfs,    ('/tmp',), True),
               (c.protocol,   (), True),
               (c.dirlist,    ('/tmp', DirListFlags.STAT), True),
               (c.sendinfo,   ('important info',), False),
               (c.prepare,    (['/tmp/foo'], PrepareFlags.STAGE), True),
               ]

  for func, args, hasReturnObject in funcspecs:
      sync (func, args, hasReturnObject)

  # Create new temp file
  f = client.File()
  status, response = f.open(smallfile, OpenFlags.NEW)

  for func, args, hasReturnObject in funcspecs:
      async(func, args, hasReturnObject)

def sync(func, args, hasReturnObject):
  status, response = func(*args)
  print status
  assert status.ok
  if hasReturnObject:
      print response
      assert response

def async(func, args, hasReturnObject):
  handler = AsyncResponseHandler()
  status = func(callback=handler, *args)
  print status
  assert status.ok
  status, response, hostlist = handler.wait()

  assert status.ok
  if response:
      assert response

  for host in hostlist:
    assert host.url
    print host.url

  if hasReturnObject:
    assert response

def test_copy_sync():
  c = client.FileSystem(SERVER_URL)
  f = client.File()
  status, response = f.open(smallfile, OpenFlags.DELETE)
  assert status.ok
  
  status, response = c.copy(smallfile, '/tmp/eggs', force=True)
  assert status.ok
  
  status, response = c.copy('/tmp/nonexistent', '/tmp/eggs')
  assert not status.ok

  try:
    os.remove('/tmp/eggs')
  except OSError, __:
    pass

def test_locate_sync():
  c = client.FileSystem(SERVER_URL)
  status, response = c.locate('/tmp', OpenFlags.REFRESH)
  assert status.ok

  for item in response:
    assert item

def test_locate_async():
  c = client.FileSystem(SERVER_URL)
  handler = AsyncResponseHandler()
  response = c.locate('/tmp', OpenFlags.REFRESH, callback=handler)

  status, response, hostlist = handler.wait()
  assert status.ok

  for item in response:
    assert item

def test_deeplocate_sync():
  c = client.FileSystem(SERVER_URL)
  status, response = c.deeplocate('/tmp', OpenFlags.REFRESH)
  assert status.ok

  for item in response:
    assert item

def test_deeplocate_async():
  c = client.FileSystem(SERVER_URL)
  handler = AsyncResponseHandler()
  response = c.deeplocate('/tmp', OpenFlags.REFRESH, callback=handler)

  status, response, hostlist = handler.wait()
  assert status.ok

  for item in response:
    assert item

def test_dirlist_sync():
  c = client.FileSystem(SERVER_URL)
  status, response = c.dirlist('/tmp', DirListFlags.STAT)
  assert status.ok

  for item in response:
    assert item.name
    print item.statinfo
    assert item.statinfo
    assert item.hostaddr
    
  status, response = c.dirlist('invalid', DirListFlags.STAT)
  assert not status.ok

def test_dirlist_async():
  c = client.FileSystem(SERVER_URL)
  handler = AsyncResponseHandler()
  status = c.dirlist('/tmp', DirListFlags.STAT, callback=handler)
  assert status.ok
  status, response, hostlist = handler.wait()
  assert status.ok

  for h in hostlist:
    print h.url

  for item in response:
    assert item.name
    print item.statinfo
    assert item.statinfo
    assert item.hostaddr

  assert hostlist

def test_query_sync():
  c = client.FileSystem(SERVER_URL)
  status, response = c.query(QueryCode.STATS, 'a')
  assert status.ok
  assert response
  print response

def test_query_async():
  c = client.FileSystem(SERVER_URL)
  handler = AsyncResponseHandler()
  status = c.query(QueryCode.STATS, 'a', callback=handler)
  assert status.ok

  status, response, hostlist = handler.wait()
  assert status.ok
  assert response
  print response
  
def test_mkdir_flags():
  c = client.FileSystem(SERVER_URL)
  status, response = c.mkdir('/tmp/dir1/dir2', MkDirFlags.MAKEPATH)
  assert status.ok
  c.rm('/tmp/dir1/dir2')
  c.rm('/tmp/dir1')
  

def test_args():
  c = client.FileSystem(url=SERVER_URL)
  assert c

  pytest.raises(TypeError, "c = client.FileSystem(foo='root://localhost')")
  pytest.raises(TypeError, "c = client.FileSystem(path='root://localhost', foo='bar')")

def test_creation():
  c = client.FileSystem(SERVER_URL)
  assert c.url is not None

def test_deletion():
  c = client.FileSystem(SERVER_URL)
  del c

  if sys.hexversion > 0x03000000:
      pytest.raises(UnboundLocalError, 'assert c')
  else:
      pytest.raises(NameError, 'assert c')





from XRootD import client
from XRootD.client.flags import OpenFlags

with client.File() as f:
  status, response = f.open('root://localhost//tmp/eggs', OpenFlags.DELETE)
  status, response = f.visa()
  print status, response



from __future__ import absolute_import, division, print_function

from .glob_funcs import glob, iglob
from .filesystem import FileSystem as FileSystem
from .file import File as File
from .url import URL as URL
from .copyprocess import CopyProcess as CopyProcess
from .env import EnvPutString as EnvPutString
from .env import EnvGetString as EnvGetString
from .env import EnvPutInt as EnvPutInt
from .env import EnvGetInt as EnvGetInt
from ._version import __version__ as __version__

import XRootD.client.finalize



"""
Read all lines from a file into a list
--------------------------------------

Produces the following output (Note how the first line is not returned with the
call to :func:`readlines()` because we ate it with the first call to
:func:`readline()`)::

  'green\n'
  ['eggs\n', 'and\n', 'spam\n']

"""
from XRootD import client

with client.File() as f:
  f.open('root://localhost//tmp/eggs')

  print '%r' % f.readline()
  print f.readlines()



"""
Read scattered data chunks in one operation
-------------------------------------------

Produces the following output::

  <buffer: 'The XROOTD project aims at giving high p', length: 40, offset: 0>
  <buffer: 'erformance, scalable  fault tolerant acc', length: 40, offset: 40>
  <buffer: 'ess to data repositories of many kinds', length: 38, offset: 80>

"""
from XRootD import client
from XRootD.client.flags import OpenFlags

with client.File() as f:
  print f.open('root://localhost//tmp/eggs', OpenFlags.UPDATE)

  f.write(r'The XROOTD project aims at giving high performance, scalable '
          +' fault tolerant access to data repositories of many kinds')

  size = f.stat()[1].size
  v = [(0, 40), (40, 40), (80, size - 80)]

  status, response = f.vector_read(chunks=v)
  print status

  for chunk in response.chunks:
    print chunk



"""
Write a chunk of data to a file
-------------------------------

Produces the following output::

  'green\neggs\nand\nspam\n'

"""
from XRootD import client
from XRootD.client.flags import OpenFlags

with client.File() as f:
  print f.open('root://localhost//tmp/eggs', OpenFlags.UPDATE)

  data = 'spam\n'
  f.write(data, offset=15)
  print f.read()



"""
Iterate over a file in chunks of the specified size
---------------------------------------------------

Produces the following output::

  'green\neggs'
  '\nand\nspam\n'

"""
from XRootD import client

with client.File() as f:
  f.open('root://localhost//tmp/eggs')

  for chunk in f.readchunks(offset=0, chunksize=10):
    print '%r' % chunk



#-------------------------------------------------------------------------------
# Copyright (c) 2012-2018 by European Organization for Nuclear Research (CERN)
# Author: Benjamin Krikler <b.krikler@cern.ch>
#-------------------------------------------------------------------------------
# This file is part of the XRootD software suite.
#
# XRootD is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# XRootD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#-------------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function

from XRootD.client.filesystem import FileSystem

import glob as gl
import os
import fnmatch
import sys
if sys.version_info[0] > 2:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


__all__ = ["glob", "iglob"]


def split_url(url):
    parsed_uri = urlparse(url)
    domain = '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)
    path = parsed_uri.path
    return domain, path


def iglob(pathname, raise_error=False):
    """
    Generates paths based on a wild-carded path, potentially via xrootd.

    Multiple wild-cards can be present in the path.

    Args:
      pathname (str):  The wild-carded path to be expanded.
      raise_error (bool):  Whether or not to let xrootd raise an error if
          there's a problem.  If False (default), and there's a problem for a
          particular directory or file, then that will simply be skipped,
          likely resulting in an empty list.

    Yields:
      (str): A single path that matches the wild-carded string
    """
    # Let normal python glob try first
    generator = gl.iglob(pathname)
    path = next(generator, None)
    if path is not None:
        yield path
        for path in generator:
            yield path
        return

    # Else try xrootd instead
    for path in xrootd_iglob(pathname, raise_error=raise_error):
        yield path


def xrootd_iglob(pathname, raise_error):
    """Handles the actual interaction with xrootd

    Provides a python generator over files that match the wild-card expression.
    """
    # Split the pathname into a directory and basename
    dirs, basename = os.path.split(pathname)

    if gl.has_magic(dirs):
        dirs = list(xrootd_iglob(dirs, raise_error))
    else:
        dirs = [dirs]

    for dirname in dirs:
        host, path = split_url(dirname)
        query = FileSystem(host)

        if not query:
            raise RuntimeError("Cannot prepare xrootd query")

        status, dirlist = query.dirlist(path)
        if status.error:
            if not raise_error:
                continue
            raise RuntimeError("'{!s}' for path '{}'".format(status, dirname))

        for entry in dirlist.dirlist:
            filename = entry.name
            if filename in [".", ".."]:
                continue
            if not fnmatch.fnmatchcase(filename, basename):
                continue
            yield os.path.join(dirname, filename)


def glob(pathname, raise_error=False):
    """
    Creates a list of paths that match pathname.

    Multiple wild-cards can be present in the path.

    Args:
      pathname (str):  The wild-carded path to be expanded.
      raise_error (bool):  Whether or not to let xrootd raise an error if
          there's a problem.  If False (default), and there's a problem for a
          particular directory or file, then that will simply be skipped,
          likely resulting in an empty list.

    Returns:
      (str): A single path that matches the wild-carded string
    """
    return list(iglob(pathname, raise_error=raise_error))



"""
Write a chunk of data to a file and test it's property system
-------------------------------------------------------------

Produces the following output::

spam\n
true\n
true\n
true\n
localhost:1094\n
root://localhost:1094//tmp/eggs\n
True\n
True\n
True\n
false\n
false\n
false\n

"""
from XRootD import client
from XRootD.client.flags import OpenFlags

with client.File() as f:
  f.open('root://localhost//tmp/eggs', OpenFlags.DELETE)

  data = 'spam\n'
  f.write(data)
  print f.read()[1]
  print f.get_property( "ReadRecovery" )
  print f.get_property( "WriteRecovery" )
  print f.get_property( "FollowRedirects" )
  print f.get_property( "DataServer" )
  print f.get_property( "LastURL" )
  print f.set_property( "ReadRecovery", "false" )
  print f.set_property( "WriteRecovery", "false" )
  print f.set_property( "FollowRedirects", "false" )
  print f.get_property( "ReadRecovery" )
  print f.get_property( "WriteRecovery" )
  print f.get_property( "FollowRedirects" )
